unit DataFile;

// TDataFile: Save database data from multiple tables in a single datafile
// Use TDatafile to store and transport database data in a single file
//
// Author: Jan Derk Stegeman
// Created: February 1st 1998
// Version: 2.0
// Last Modified: June 1998

// Set if registered version or not
//{$DEFINE REGISTERED}

interface

uses sysutils, bde, db, dbtables, classes, dialogs;

// Constants
const
  DataFileID = '#$%datafile';   // Datafile identifier
  FieldID    = '#$%field';      // Field identifier
  RecordID   = '#$%record';     // Record identifier
  NearlyZero = 1e-15;

// Type declarations
type
  // Event declarations
  TOnReadFieldEvent = procedure(Sender:         TObject;
                                DataBaseName:   string;
                                TableName:      string;
                                var FieldName:  string;
                                var FieldType:  TFieldType;
                                var ReadField:  boolean) of object;

  TOnReadRecordEvent = procedure(Sender:           TObject;
                                 var DataBaseName: string;
                                 var TableName:    string;
                                 var ReadRecord:   boolean) of object;

  TOnUpdateRecordEvent = procedure(Sender:           TObject;
                                   Destination:      TTable;
                                   var UpdateRecord: boolean) of object;

  // Exception classes declarations
  EDataFileError = class(Exception);

  // Loadmode declaration
  TLoadMode = (lmAppend, lmAppendUpdate, lmUpdate, lmDelete);
  {lmAppend: Appends records from datafile to destination table.

   lmUpdate: Replace records in table with matching records from datafile.
      The table must have an index defined to match records.

   lmAppendUpdate: If a matching record exists in the table, replace it.
     Otherwise, append records to the table. The destination table must have
     an index defined to match records.
   lmDelete: Delete records in destination table that match with records
     from the datafile}

  TDataFileOption = (dfCheckSignature, dfCheckVersion, dfLoadBLOBFields,
                     dfSaveBLOBFields);
  TDataFileOptions = set of TDataFileOption;

  // File header type definition
  TDataHeader = record
    ID:           string;  // Datafile ID string
    Signature:    string;  // ID string
    Version:      double;  // Version number
  end;


  // Record header type definition
  TRecordHeader = record
    ID:              string;  // Standard record ID
    DataBaseName:    string;  // Database name
    TableName:       string;  // Name of table in database
    IndexFieldNames: string;  // Fields that make up index
  end;


  // Field header type definition
  TFieldHeader = record
    ID:           string;     // Standard field ID
    FieldName:    string;     // Name of field
    FieldType:    TFieldType; // Type of field
    DataSize:     longint;    // Size of data
    IsNull:       boolean;
  end;


  // TDataFile type definition
  TDataFile = class(TComponent)
  private
    // events
    fOnReadField:     TOnReadFieldEvent;
    fOnReadRecord:    TOnReadRecordEvent;
    fOnUpdateRecord:  TOnUpdateRecordEvent;

    // Properties
    fAbortOnKeyViol: boolean;
    fKeyViolCount:   LongInt;     // Number of key violations
    fOptions:        TDataFileOptions;
    fLoadMode:       TLoadMode;
    fSource:         TTable;
    fSignature:      string;      // Signature of DataFile
    fVersion:        double;      // Version of DataFile
    fFieldCount:     LongInt;     // Number of fields read or written on current record
    fRecordCount:    LongInt;     // Number of records read or written

    // Other declarations
    DataHeader:  TDataHeader;  // Header of DataFile containing general info
    FieldHeader: TFieldHeader; // Header of each field containing field info
    Destination: TTable;       // Destination table

    // Private functions and procedures
    function CheckFileID(Stream: TStream): boolean;
    function LastField(Stream: TStream): boolean;
    function ReadCharacter(Stream: TStream): Char;
    function ReadString(Stream: TStream): string;
    procedure ReadField(Stream: TStream);
    procedure ReadFieldHeader(Stream: TStream);
    procedure ReadDataHeader(Stream: TStream);
    procedure ReadRecord(Stream: TStream);
    procedure SetVersion(AVersion: double);
    procedure SetSignature(ASignature: string);
    procedure SkipField(Stream: TStream);
    procedure WriteString(Stream: TStream; S: string);
    procedure WriteField(Stream: TStream; Field: TField);
    procedure WriteFieldHeader(Stream: TStream; Field: TField);
    procedure WriteHeader(Stream: TStream);
    procedure WriteRecord(Stream: TStream);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    // properties
    property FieldCount:   LongInt read fFieldCount;
    property KeyViolCount: LongInt read fKeyViolCount;
    property RecordCount:  LongInt read fRecordCount;

    // Public procedures
    function GetFileVersion(FileName: string): double;
    function GetFileSignature(FileName: string): string;
    function IsDataFile(FileName: string): boolean;
    procedure AppendRecord(const FileName: string);
    procedure AppendRecordToStream(Stream: TStream);
    procedure AppendTable(const FileName: string);
    procedure AppendTableToStream(Stream: TStream);
    procedure Load(const FileName: string);
    procedure LoadFromStream(Stream: TStream);

  published
    // Properties
    property AbortOnKeyViol: boolean read fAbortOnKeyViol write fAbortOnKeyViol;
    property LoadMode:  TLoadMode    read fLoadMode       write fLoadMode;
    property Options:   TDataFileOptions
                                     read fOptions        write fOptions;
    property Signature: string       read fSignature      write SetSignature;
    property Source:    TTable       read fSource         write fSource;
    property Version:   double       read fVersion        write SetVersion;

    // Events
    property OnReadField: TOnReadFieldEvent
      read fOnReadField write fOnReadField;
    property OnReadRecord: TOnReadRecordEvent
      read fOnReadRecord write fOnReadRecord;
    property OnUpdateRecord: TOnUpdateRecordEvent
      read fOnUpdateRecord write fOnUpdateRecord;
  end;

procedure Register;


implementation

procedure Register;
begin
  RegisterComponents('Data Access', [TDataFile]);
end;


function TDataFile.CheckFileID(Stream: TStream): boolean;
var
  i: integer;
  S: string;
  C: Char;
  StorePosition: LongInt;
begin
  Result := False;

  // Leave function if remaining stream length is to short
  if (Stream.Size - Stream.Position) < Length(DataFileID) then Exit;

  StorePosition := Stream.Position;
  try
    S := '';
    C := ReadCharacter(Stream);

    for i := 1 to Length(DataFileID) do
    begin
      S := S + C;
      C := ReadCharacter(Stream);
    end;
  finally
    Stream.Position := StorePosition;
  end;

  if S = DataFileID then Result := True;
end;


// Check if current field is last field of record
function TDataFile.LastField(Stream: TStream): boolean;
var
  i: integer;
  S: string;
  C: Char;
  StorePosition: LongInt;
begin
  LastField := True;

  // Leave function if remaining stream length is to short
  if (Stream.Size - Stream.Position) < Length(FieldID) then Exit;

  StorePosition := Stream.Position;
  try
    S := '';
    C := ReadCharacter(Stream);

    for i := 1 to Length(FieldID) do
    begin
      S := S + C;
      C := ReadCharacter(Stream);
    end;
  finally
    Stream.Position := StorePosition;
  end;

  if S = FieldID then LastField := False
end;


// Read one character from stream
function TDataFile.ReadCharacter(Stream: TStream): Char;
begin
  if Stream.Position = Stream.Size then
  MessageDLG('Error: Unexpected end of stream encountered',
             mtError, [mbOK], 0);
  Stream.Read(Result, 1);
end;


// Read a long string from stream
function TDataFile.ReadString(Stream: TStream): string;
var
  C: Char;
  S: string;
begin
  S := '';
  // Read first character
  C := ReadCharacter(Stream);
  while C <> #0 do
  begin
    S := S + C;
    C := ReadCharacter(Stream);
  end;
  Result := S;
end;


// Read a field from stream
procedure TDataFile.ReadField(Stream: TStream);
var
  Field:      TField;
  Buffer:     PChar;
  BlobStream: TBlobstream;
  ReadField:  Boolean;
begin
  with FieldHeader do
  begin
    ReadFieldHeader(Stream);

    // Default is: read every field
    ReadField := True;

    // Check if BlobFields should be read by default
    if not (dfLoadBlobFields in Options) and
       (FieldType in [ftBlob, ftMemo, ftGraphic, ftFmtMemo,
                     ftParadoxOLE, ftDBaseole, ftTypedBinary]) then
      ReadField := False;

    // Call OnReadField event
    if Assigned(OnReadField) then OnReadField(Self,
                                              Destination.DataBaseName,
                                              Destination.TableName,
                                              FieldName,
                                              FieldType,
                                              ReadField);

    // Read field data
    if not ReadField then
    begin
      // Skip field if ReadField = False
      Stream.Position := Stream.Position + DataSize;
      Exit;
    end;

    // Error checking
    if ID <> FieldID then
      raise EDataFileError.Create('Unable to read datafile.');

    // Find field in target table
    Field := Destination.FindField(FieldName);

    if Field = nil then
      raise EDataFileError.Create('Field not found.'#13#10 +
        'Field named ''' + FieldName + ''' could not be found'#13#10 +
        'in ''' + Destination.TableName + ''' table.');

    if FieldType <> Field.DataType then
      raise EDataFileError.Create('Unexpected field type.'#13#10 +
        'Fieldtype in DataFile differs from fieldtype in ''' +
        Destination.TableName + ''' table.');

    // Skip field if indexfield or not depending on table state
    if (LoadMode in [lmUpdate, lmAppendUpdate]) and
       (Destination.State = dsSetKey)           and
       (not Field.IsIndexField)                 then
    begin
      Stream.Position := Stream.Position + DataSize;
      exit;
    end;

    if (LoadMode = lmUpdate)                     and
       (Destination.State in [dsEdit, dsInsert]) and
       Field.IsIndexField                        then
    begin
      Stream.Position := Stream.Position + DataSize;
      exit;
    end;


    // Read field data
    if (DataSize > 0)  and (not IsNull) then
    begin
      if Field.isBlob then
      begin
        BlobStream := TBlobStream.Create(Field as TBlobField, bmWrite);
        try
          BlobStream.Copyfrom(Stream, DataSize)
        finally
          BlobStream.Free;
        end;
      end  //if Field.isBlob then
      else
      begin // Not BLOB field
        // Load field data
        GetMem(Buffer, DataSize);
        try
          Stream.ReadBuffer(Buffer^, DataSize);
          Field.SetData(Buffer);
        finally
          FreeMem(Buffer, DataSize);
        end;
      end;
    end  // if (DataSize > 0)  and (not IsNull) then
    else if (DataSize > 0) and IsNull then
    begin
      Stream.Position := Stream.Position + DataSize;
      Field.Clear; // Clear field
    end
    else Field.Clear; // Clear field if datasize is zero

    // Increase FieldCounter if field read succesfully
    Inc(fFieldCount);
  end; // with FieldHeader do
end;


procedure TDataFile.ReadFieldHeader(Stream: TStream);
begin
  with FieldHeader do
  begin
    // Read field header
    ID        := ReadString(Stream);
    FieldName := ReadString(Stream);
    Stream.ReadBuffer(FieldType,    SizeOf(TFieldType));
    Stream.ReadBuffer(IsNull,       SizeOf(IsNull));
    Stream.ReadBuffer(DataSize,     SizeOf(DataSize));
  end;
end;


procedure TDataFile.ReadDataHeader(Stream: TStream);
begin
  if not CheckFileID(Stream) then
    raise EDataFileError.Create('File is not a ''' + Signature + ''' file');

  // Read file header
  DataHeader.ID := ReadString(Stream);
  if DataHeader.ID <> DataFileID then
    raise EDataFileError.Create('File is not a ''' + Signature + ''' file');

  DataHeader.Signature := ReadString(Stream);
  Stream.ReadBuffer(DataHeader.Version, SizeOf(DataHeader.Version));

  // Error checking
  if (dfCheckSignature in fOptions)      and
     (DataHeader.Signature <> Signature) then
    raise EDataFileError.Create('File is not a ''' + Signature + ''' file');

  if (dfCheckVersion in fOptions)    and
     (abs(DataHeader.Version - Version) > NearlyZero) then
  begin
    if DataHeader.Version < Version then
      raise EDataFileError.Create('Old file version.'#13#10 +
        'File version: '     + FloatToStr(DataHeader.Version) + #13#10 +
        'Expected version: ' + FloatToStr(Version))
    else
      raise EDataFileError.Create('File version newer than expected.'#13#10 +
        'File version: '     + FloatToStr(DataHeader.Version) + #13#10 +
        'Expected version: ' + FloatToStr(Version));
  end;
end;


procedure TDataFile.ReadRecord(Stream: TStream);
var
  ReadRecord:    boolean;
  UpdateRecord:  boolean;
  StorePosition: LongInt;       // Stream position storage
  RecordHeader:  TRecordHeader; // Stream header record
begin
  // Initializations
  fFieldCount := 0;

  // Read and error check record header
  with RecordHeader do
  begin
    ID              := ReadString(Stream);
    DataBaseName    := ReadString(Stream);
    TableName       := ReadString(Stream);
    IndexFieldNames := ReadString(Stream);

    // Default is: read all records
    ReadRecord :=  True;

    // Call OnReadField event
    if Assigned(OnReadRecord) then OnReadRecord(Self,
                                                DataBaseName,
                                                TableName,
                                                ReadRecord);
  end;

  if not ReadRecord then
  begin
    // Goto end of record and exit ReadRecord procedure
    while not LastField(Stream) do SkipField(Stream);
    Exit;
  end;

  // Error checking
  if RecordHeader.ID <> RecordID then
    raise EDataFileError.Create('Unable to read datafile.');

  if RecordHeader.DatabaseName <> Destination.DataBaseName then
  begin
    if Destination.Active then Destination.Close;
    Destination.DatabaseName := RecordHeader.DatabaseName;
  end;

  if RecordHeader.TableName <> Destination.TableName then
  begin
    if Destination.Active then Destination.Close;
    Destination.TableName := RecordHeader.TableName;
  end;

  if RecordHeader.IndexFieldNames <> Destination.IndexFieldNames then
  begin
    if Destination.Active then Destination.Close;
    Destination.IndexFieldNames := RecordHeader.IndexFieldNames;
  end;

  // Open detination table
  Destination.Open;


  // Store begin of record stream position
  StorePosition := Stream.Position;

  // Find Key
  Destination.SetKey;

  // Set keyfields
  while (fFieldCount < Destination.IndexFieldCount) and
        (not LastField(Stream)) do ReadField(Stream);

  // Check if all key fields have been set
  if FieldCount < Destination.IndexFieldCount then
    raise EDataFileError.Create('Not all key fields were available in datafile.');

  if Destination.GotoKey then
  begin
    case FLoadMode of
      lmUpdate, lmAppendUpdate:
      begin
        // Update record
        UpdateRecord := True;

        // Call OnUpdateRecord event
        if Assigned(OnUpdateRecord) then OnUpdateRecord(Self,
                                                        Destination,
                                                        UpdateRecord);
        if UpdateRecord then
        begin
          try
            Destination.Edit;
          except
            // Cancel record if destination table can not get into edit mode
            Destination.Cancel;
            while not LastField(Stream) do SkipField(Stream);
          end;

          // Read table fields
          try
            while not LastField(Stream) do ReadField(Stream);
          except
            // Cancel record changes if anything goes wrong
            Destination.Cancel;

            // Reraise exception
            raise;
          end;
        end
        else while not LastField(Stream) do SkipField(Stream);
      end;

      lmAppend:
      begin
        Inc(fKeyViolCount);
        if fAbortOnKeyViol then raise EDataFileError.Create('Key violation.')
        else while not LastField(Stream) do SkipField(Stream);
      end;

      lmDelete:
      begin
        // Delete record
        Destination.Delete;
        while not LastField(Stream) do SkipField(Stream);
      end;
    end; // case FLoadMode of
  end

  else // if Destination.GotoKey then
  begin
    case FLoadMode of
      lmUpdate, lmDelete: while not LastField(Stream) do SkipField(Stream);
      lmAppend, lmAppendUpdate:
      begin
        // Append field
        Stream.Position := StorePosition;
        Destination.Append;
        try
          // Read table fields
          while not LastField(Stream) do ReadField(Stream);
        except
          // Cancel record changes if anything goes wrong
          Destination.Cancel;

          // Reraise exception
          raise;
        end;
      end;
    end;
  end; // if Destination.GotoKey then

  // Post table updates
  try
    if Destination.State in [dsEdit, dsInsert] then Destination.Post;
  except
    Destination.Cancel;
  end;
end;


procedure TDataFile.SetSignature(ASignature: string);
begin
  if ASignature <> FSignature then
    if ASignature = '' then
      MessageDLG('Signature property can not be empty.', mtInformation,
                 [mbok], 0)
    else FSignature := ASignature;
end;


procedure TDataFile.SetVersion(AVersion: double);
begin
  if abs(AVersion - FVersion) > NearlyZero then
    if AVersion >= 0 then FVersion := AVersion
    else MessageDLG('Version should be larger than zero', mtInformation, [mbok], 0);
end;




// Read field but don't use it
procedure TDataFile.SkipField(Stream: TStream);
begin
  ReadFieldHeader(Stream);

  // Error checking
  if FieldHeader.ID <> FieldID then
    raise EDataFileError.Create('Unable to read datafile.');

  Stream.Position := Stream.Position + FieldHeader.DataSize;
end;


// Write a string to stream
procedure TDataFile.WriteString(Stream: TStream; s: string);
var
  c: char;
begin
  c := #0;
  if s = '' then Stream.Write(c, 1)
  else
  // String size is string length plus termination null character
  Stream.WriteBuffer(Pointer(S)^, Length(S) + 1);
end;


procedure TDataFile.WriteField(Stream: TStream; Field: TField);
var
  Buffer:     PChar;
  BlobStream: TBlobStream;
begin
  if Field.isBlob then
  begin
    if dfSaveBLOBFields in Options then
    begin
      WriteFieldHeader(Stream, Field);
      BlobStream := TBlobStream.Create(Field as TBlobField, bmRead);
      try
        if BlobStream.Size > 0 then Stream.Copyfrom(BlobStream, BlobStream.Size);
      finally
        BlobStream.Free;
      end;
    end; // if dfLoadBLOBFields in Options then
  end  // if Field.isBlob then
  else
  begin // Load no Blob field
    WriteFieldHeader(Stream, Field);

    if Field.DataSize > 0 then
    begin
      // Save actual field data
      GetMem(Buffer, Field.DataSize);
      try
        Field.GetData(Buffer);
        Stream.WriteBuffer(Buffer^, Field.DataSize);
      finally
        FreeMem(Buffer, Field.DataSize);
      end;
    end;
  end;
end;


procedure TDataFile.WriteFieldHeader(Stream: TStream; Field: TField);
begin
  with FieldHeader do
  begin
    ID           := FieldID;
    FieldName    := Field.FieldName;
    FieldType    := Field.DataType;
    IsNull       := Field.IsNull;
    if Field.isBlob then DataSize := (Field as TBlobField).BlobSize
    else DataSize := Field.DataSize;

    WriteString(Stream, ID);
    WriteString(Stream, FieldName);
    Stream.WriteBuffer(FieldType,    SizeOf(TFieldType));
    Stream.WriteBuffer(IsNull,       SizeOf(IsNull));
    Stream.WriteBuffer(DataSize,     SizeOf(DataSize));
  end;
end;


procedure TDataFile.WriteHeader(Stream: TStream);
begin
  {$IFNDEF REGISTERED}
  MessageDLG('Thank you for evaluating TDataFile.'#13#10 +
             #13#10 +
             'by Jan Derk Stegeman'#13#10 +
             'UNREGISTERED VERSION!'#13#10 +
             #13#10 +
             'Read ''readme.txt'' for ordering information.',
             mtInformation, [mbok], 0);
  {$ENDIF}

  // Set file header values
  DataHeader.ID        := DataFileID;
  DataHeader.Signature := Signature;
  if DataHeader.Signature = '' then
    raise EDataFileError.Create('Signature property is empty.');

  DataHeader.Version := Version;
  if DataHeader.Version < 0  then
    raise EDataFileError.Create('Incorrect datafile version number.'#13#10 +
                                'Version number must be positive.');

  // Write DataHeader to file
  WriteString(Stream, DataHeader.ID);
  WriteString(Stream, DataHeader.Signature);
  Stream.WriteBuffer(DataHeader.Version, SizeOf(DataHeader.Version));
end;


procedure TDataFile.WriteRecord(Stream: TStream);
var
  i: integer;
  s: string;
  RecordHeader: TRecordHeader;
begin
  with RecordHeader do
  begin
    ID              := RecordID;
    DataBaseName    := Source.DatabaseName;
    TableName       := Source.TableName;

    WriteString(Stream, ID);
    WriteString(Stream, DataBaseName);
    WriteString(Stream, TableName);

    // Write indexFieldNames
    if Source.IndexFieldNames <> '' then
      WriteString(Stream, Source.IndexFieldNames)
    else
    begin
      s := '';
      for i := 0 to Source.IndexFieldCount - 1 do
      begin
        if i > 0 then s := s + ';';
        s := s + Source.IndexFields[i].FieldName;
      end;
      WriteString(Stream, s);
    end;

  end;

  if not Source.Active then Source.Open;

  // First save index fields
  for i := 0 to Source.IndexFieldCount - 1 do
    WriteField(Stream, Source.IndexFields[i]);

  // Save the other non-index fields
  for i := 0 to Source.FieldCount - 1 do
    if not Source.Fields[i].IsIndexField then
      WriteField(Stream, Source.Fields[i]);
end;


//-----------------------------------
// Public methods
//-----------------------------------
constructor TDataFile.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  // Set default properties values
  // Properties which are also available in TBatchMove have the same defaults
  fAbortOnKeyViol := True;
  fFieldCount     := 0;
  fKeyViolCount   := 0;
  fLoadMode       := lmAppend;
  fOptions        := [dfCheckVersion, dfCheckSignature, dfLoadBLOBFields,
                      dfSaveBLOBFields];
  fRecordCount    := 0;
  fSignature      := 'DataFileSignature';
  fVersion        := 1;
end;


destructor TDataFile.Destroy;
begin
  Inherited Destroy;
end;


function TDataFile.GetFileSignature(FileName: string): string;
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    ReadDataHeader(Stream);
    Result := DataHeader.Signature;
  finally
    Stream.Free;
  end;
end;


function TDataFile.GetFileVersion(FileName: string): double;
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    ReadDataHeader(Stream);
    Result := DataHeader.Version;
  finally
    Stream.Free;
  end;
end;


function TDataFile.IsDataFile(FileName: string): boolean;
var
  Stream: TStream;
begin
  Result := False;

  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    if CheckFileID(Stream) then Result := True;
  finally
    Stream.Free;
  end;
end;


procedure TDataFile.AppendRecord(const FileName: string);
var
  Stream: TStream;
begin
  if FileExists(FileName) then
    Stream := TFileStream.Create(FileName, fmOpenReadWrite)
  else Stream := TFileStream.Create(FileName, fmCreate);

  try
    AppendRecordToStream(Stream);
  finally
    Stream.Free;
  end;
end;


procedure TDataFile.AppendRecordToStream(Stream: TStream);
// Append current record to stream
var
  StoreActive: boolean;
begin
  // Source property should not be empty
  if Source = nil then raise EDataFileError.Create('Source property is empty.');

  StoreActive := Source.Active;

  // Write DataHeader if stream is empty
  // else check if stream is of correct type
  if Stream.Size = 0 then WriteHeader(Stream)
  else
  begin
    ReadDataHeader(Stream);
    Stream.Position := Stream.Size;
  end;

  WriteRecord(Stream);

  if not StoreActive then Source.Close;
end;


procedure TDataFile.AppendTable(const FileName: string);
var
  Stream: TStream;
begin
  if FileExists(FileName) then
    Stream := TFileStream.Create(FileName, fmOpenReadWrite)
  else Stream := TFileStream.Create(FileName, fmCreate);

  try
    AppendTableToStream(Stream);
  finally
    Stream.Free;
  end;
end;


// Append all records from source to datafile
procedure TDataFile.AppendTableToStream(Stream: TStream);
var
 Bookmark: TBookmark;
begin
  if Source = nil then raise EDataFileError.Create('Source property is empty.');

  // Write DataHeader if stream is empty
  // else check if stream is of correct type
  if Stream.Size = 0 then WriteHeader(Stream)
  else
  begin
    ReadDataHeader(Stream);
    Stream.Position := Stream.Size;
  end;

  Bookmark := Source.GetBookmark;
  Source.DisableControls;
  try
    Source.First;
    // Iterate through each record in table
    while not Source.EOF do
    begin
      WriteRecord(Stream);
      Source.Next;
    end;
  finally
    Source.GotoBookmark(Bookmark);
    Source.EnableControls;
    Source.FreeBookmark(Bookmark);
  end;
end;


procedure TDataFile.Load(const FileName: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;


procedure TDataFile.LoadFromStream(Stream: TStream);
begin
  {$IFNDEF REGISTERED}
  MessageDLG('Thank you for evaluating TDataFile.'#13#10 +
             #13#10 +
             'by Jan Derk Stegeman'#13#10 +
             'UNREGISTERED VERSION!'#13#10 +
             #13#10 +
             'Read ''readme.txt'' for ordering information.',
             mtInformation, [mbok], 0);
  {$ENDIF}

  Destination := TTable.Create(Self);
  try
    // Initializations
    fRecordCount  := 0;
    fKeyViolCount := 0;

    ReadDataHeader(Stream);

    while stream.Position < Stream.Size do
    begin
      ReadRecord(Stream);
      Inc(fRecordCount);
    end;
  finally
    Destination.Free;
  end;
end;

end.
